#include "Squinewave.hpp"

#include <stdint.h>
#include <memory.h>


// Module constants
namespace {
	static const double Pi = 3.1415926535897932384626433832795;  // would it sound better if i enter more decimals here...? ;)

	template<class N, class M>
	inline N Clamp(N x, M minval, M maxval) {
		// Returns maxval on Inf and NaN.
		return (x >= minval && x <= maxval) ? x : (x < minval) ? minval : maxval;
	}
}

// ==================================================================

static inline int32_t find_sync(const double* sync_sig, uint32_t first, uint32_t last)
{
	if (sync_sig == 0)
		return -1;

	for (uint32_t i = first; i < last; ++i) {
		if (sync_sig[i] >= 1.0)
			return i;
	}
	return -1;
}

/* ================================================================== */

void
Squinewave::hardsync_init(const double freq)
{
	if (hardsync_phase)
		return;

	// If we're in last flat part, we're just done now
	if (warped_phase == 2.0) {
		phase = 2.0;
		return;
	}

	if (freq > Max_Warp_Freq)
		return;

	hardsync_inc = (Pi / Min_Sweep);
	hardsync_phase = hardsync_inc * 0.5; 
}


// ==================================================================


void Squinewave::generate(double* output, const long count)
{
	int32_t sync = -1;
	if (sync_sig_in)
		sync = find_sync(sync_sig_in, 0, count);
	if (sync_sig_out)
		memset(sync_sig_out, 0, count * sizeof(double));

	for (int n = 0; n < count; n++)
	{
		double freq = fmax(freq_param.get_next(n), 0.0);

		if (sync == n)
			hardsync_init(freq);

		if (hardsync_phase) {
			const double syncsweep = 0.5 * (1.0 - cos(hardsync_phase));
			freq += syncsweep * ((2.0 * Max_Warp_Freq) - freq);
			hardsync_phase += hardsync_inc;
			if (hardsync_phase > Pi) {
				hardsync_phase = Pi;
				hardsync_inc = 0.0;
			}
		}

		const double phase_inc = freq * Maxphase_By_sr;

		// Pure sine if freq > sr/(2*Min_Sweep)
		if (freq >= Max_Warp_Freq)
		{
			// Continue from warped
			output[n] = cos(Pi * warped_phase);
			phase = warped_phase;
			warped_phase += phase_inc;
		}
		else
		{
			const double min_sweep = phase_inc * Min_Sweep;
			const double skew = 1.0 - Clamp(skew_param.get_next(n), -1.0, 1.0);
			const double clip = 1.0 - Clamp(clip_param.get_next(n), 0.0, 1.0);
			const double midpoint = Clamp(skew, min_sweep, 2.0 - min_sweep);

			// 1st half: Sweep down to cos(warped_phase <= Pi) then flat -1 until phase >= midpoint
			if (warped_phase < 1.0 || (warped_phase == 1.0 && phase < midpoint))
			{
				if (warped_phase < 1.0)
				{
					const double sweep_length = fmax(clip * midpoint, min_sweep);

					output[n] = cos(Pi * warped_phase);
					warped_phase += fmin(phase_inc / sweep_length, Max_Warp);

					// Handle fractional warped_phase overshoot after sweep ends
					if (warped_phase > 1.0) {
						/* Tricky here: phase and warped may disagree where we are in waveform (due to FM + skew/clip changes).
						 * Warped dominates to keep waveform stable, waveform (flat part) decides where we are. 
						 */
						const double flat_length = midpoint - sweep_length;
						// warp overshoot scaled to main phase rate
						const double phase_overshoot = (warped_phase - 1.0) * sweep_length;

						// phase matches shape
						phase = midpoint - flat_length + phase_overshoot - phase_inc;

						// Flat if next samp still not at midpoint
						if (flat_length >= phase_overshoot) {
							warped_phase = 1.0;
							// phase may be > midpoint here (which means actually no flat part),
							// if so it will be corrected in 2nd half (since warped == 1.0)
						}
						else {
							const double next_sweep_length = fmax(clip * (2.0 - midpoint), min_sweep);
							warped_phase = 1.0 + (phase_overshoot - flat_length) / next_sweep_length;
						}
					}
				}
				else {
					// flat up to midpoint
					output[n] = -1.0;
					warped_phase = 1.0;
				}
			}
			// 2nd half: Sweep up to cos(warped_phase <= 2.Pi) then flat +1 until phase >= 2
			else {
				if (warped_phase < 2.0)
				{
					const double sweep_length = fmax(clip * (2.0 - midpoint), min_sweep);
					if (warped_phase == 1.0) {
						// warped_phase overshoot after flat part
						warped_phase = 1.0 + fmin( fmin(phase - midpoint, phase_inc) / sweep_length, Max_Warp);
					}
					output[n] = cos(Pi * warped_phase);
					warped_phase += fmin(phase_inc / sweep_length, Max_Warp);

					if (warped_phase > 2.0) {
						const double flat_length = 2.0 - (midpoint + sweep_length);
						const double phase_overshoot = (warped_phase - 2.0) * sweep_length;

						phase = 2.0 - flat_length + phase_overshoot - phase_inc;

						if (flat_length >= phase_overshoot) {
							warped_phase = 2.0;
						}
						else {
							const double next_sweep_length = fmax(clip * midpoint, min_sweep);
							warped_phase = 2.0 + (phase_overshoot - flat_length) / next_sweep_length;
						}
					}
				}
				else {
					output[n] = 1.0;
					warped_phase = 2.0;
				}
			}
		}

		phase += phase_inc;
		if (warped_phase >= 2.0 && phase >= 2.0)
		{
			if (hardsync_phase) {
				warped_phase = phase = 0.0;
				hardsync_phase = hardsync_inc = 0.0;

				sync = find_sync(sync_sig_in, n + 1, count);
			}
			else {
				phase -= 2.0;
				if (phase > phase_inc) {
					// wild aliasing freq - just reset
					phase = phase_inc * 0.5;
				}
				if (freq < Max_Warp_Freq) {
					const double min_sweep = phase_inc * Min_Sweep;
					const double skew = 1.0 - Clamp(skew_param.get_current(), -1.0, 1.0);
					const double clip = 1.0 - Clamp(clip_param.get_current(), 0.0, 1.0);
					const double midpoint = Clamp(skew, min_sweep, 2.0 - min_sweep);
					const double next_sweep_length = fmax(clip * midpoint, min_sweep);
					warped_phase = fmin(phase / next_sweep_length, Max_Warp);
				}
				else
					warped_phase = phase;
			}

			if (sync_sig_out)
				sync_sig_out[n] = 1.0;
		}
	}
}

// ==================================================================

