/**
	@file
	squinewave~: sine-saw-square-pulse oscillator for Max
	original by: rasmus.ekman@abc.se
	@ingroup oscillators	
*/

#include "maxcpp6.h"    // standard Max include, always required (except in Jitter)

#include "Squinewave.hpp"



// Inlets: Freq, clip, skew, sync
// Outlet: sound, sync
class MspSquinewave : public MspCpp6<MspSquinewave>
{
public:
	enum {
		Freq_Inlet,
		Clip_Inlet,
		Skew_Inlet,
		Sync_Inlet,
		Nof_Inlets
	};

	MspSquinewave(t_symbol * sym, long ac, t_atom * av);

	~MspSquinewave();

	void perform(double **ins, long numins, double **outs, long numouts, long sampleframes);

	void assist(void *b, long io, long index, char *dst);

	void float_inlets(long index, double val);

	// Accept bang on sync inlet
	void bang(long inlet);

private:
	Squinewave synth;

	// Switch between signal and single-float input
	// if sig_in is all 0, use val (single)
	double* check_input(double* sig_in, long count);

};



// ==================================================================
// Constructor

MspSquinewave::MspSquinewave(t_symbol * sym, long ac, t_atom * av)
{
	// 4 inputs: Freq, Clip, Skew, Sync
	// 2 outputs: Audio and sync out
	setupIO(Nof_Inlets, 2);
	synth.framerate = sys_getsr();
	// 1 input argument: Min_Sweep
	if (ac) {
		if (av->a_type != A_LONG) {
			object_post(nullptr, "squine~ min sweep must be integer (4-100)");
		}
		else {
			int minsweep = int(av->a_w.w_long);
			// 0: rand range 10-20
			if (minsweep == 0) {
				minsweep = 10 + (rand() % 10);
			}
			if (minsweep < 4 || minsweep > 100) {
				object_post(nullptr, "squine~ min sweep range: 4-100");
				minsweep = (minsweep < 4) ? 4 : 100;
			}

			synth.Min_Sweep = minsweep;
#if _DEBUG
			object_post(nullptr, "squine~ min sweep: %i samples", minsweep);
#endif
		}
	}

#if _DEBUG
	object_post(nullptr, "squine~dbg " __TIME__ " created");
#endif
}

MspSquinewave::~MspSquinewave() {
#if _DEBUG
	object_post(nullptr, "squine~dbg freed");
#endif
}

// ==================================================================
// Help strings

void
MspSquinewave::assist(void *b, long io, long index, char *dst)
{
	// io 1 = inlet, 2 = outlet
	switch (io)
	{
	case ASSIST_INLET:
		switch (index) {
		case Freq_Inlet: strncpy_zero(dst, "Frequency", 100); break;
		case Clip_Inlet: strncpy_zero(dst, "Clip (0-1) - squareness", 100); break;
		case Skew_Inlet: strncpy_zero(dst, "Skew (+/-1) - symmetry offset", 100); break;
		case Sync_Inlet: strncpy_zero(dst, "Sync (0 or 1)", 100); break;
		}
		break;
	case ASSIST_OUTLET:
		if (index == 0)
			strncpy_zero(dst, "Audio", 512);
		else
			strncpy_zero(dst, "Sync out", 512);
		break;
	}

}

// ==================================================================
// Inlets

void
MspSquinewave::float_inlets(long index, double val)
{
	switch (index) {
		case Freq_Inlet: synth.freq_in = val; break;
		case Clip_Inlet: synth.clip_in = val; break;
		case Skew_Inlet: synth.skew_in = val; break;
	}
}

void
MspSquinewave::bang(long inlet)
{
	if (inlet == Sync_Inlet)
		synth.hardsync_now();
}

double*
MspSquinewave::check_input(double* sig_in, long count)
{
	for (int i = 0; i < count; ++i) {
		if (sig_in[i]) {
			return sig_in;
		}
	}
	return nullptr;
}

// ==================================================================
// Synth

void
MspSquinewave::perform(double **ins, long numins, double **outs, long numouts, long sampleframes)
{
	// This just shouldn't happen.
	if (numins <= Sync_Inlet) {
		static bool error_sent = false;
		if (error_sent)
			object_post(nullptr, "Error: Invalid inputs %i", numins);
			return;
	}
	// Check which signal~ inputs have values (no way to know if a cable is connected?)
	// Maybe stoopid "optimization" not to walk entire array?
	synth.freq_sig_in = check_input(ins[Freq_Inlet], 4);
	synth.clip_sig_in = check_input(ins[Clip_Inlet], 4);
	synth.skew_sig_in = check_input(ins[Skew_Inlet], 4);
	// This check done in generate()
	synth.sync_sig_in = ins[Sync_Inlet];

	double* output = outs[0];
	synth.sync_sig_out = outs[1];

	synth.generate(output, sampleframes);
}

// ==================================================================
// DLL main


C74_EXPORT int main(void)
{
	// create a class with the given name
	MspSquinewave::makeMaxClass("squinewave~");

	REGISTER_METHOD_ASSIST(MspSquinewave, assist);
	REGISTER_INLET_FLOAT(MspSquinewave, float_inlets);
	REGISTER_METHOD(MspSquinewave, bang);
}

// ==================================================================
// #EOF

